# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst._parser.entrypoints import is_native
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class ClassDefCreationTest(CSTNodeTest):
    @data_provider(
        (
            # Simple classdef
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"), cst.SimpleStatementSuite((cst.Pass(),))
                ),
                "code": "class Foo: pass\n",
                "expected_position": CodeRange((1, 0), (1, 15)),
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(): pass\n",
            },
            # Positional arguments render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    bases=(cst.Arg(cst.Name("obj")),),
                ),
                "code": "class Foo(obj): pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    bases=(
                        cst.Arg(cst.Name("Bar")),
                        cst.Arg(cst.Name("Baz")),
                        cst.Arg(cst.Name("object")),
                    ),
                ),
                "code": "class Foo(Bar, Baz, object): pass\n",
            },
            # Keyword arguments render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    keywords=(
                        cst.Arg(keyword=cst.Name("metaclass"), value=cst.Name("Bar")),
                    ),
                ),
                "code": "class Foo(metaclass = Bar): pass\n",
                "expected_position": CodeRange((1, 0), (1, 32)),
            },
            # Iterator expansion render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    bases=(cst.Arg(star="*", value=cst.Name("one")),),
                ),
                "code": "class Foo(*one): pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    bases=(
                        cst.Arg(star="*", value=cst.Name("one")),
                        cst.Arg(star="*", value=cst.Name("two")),
                        cst.Arg(star="*", value=cst.Name("three")),
                    ),
                ),
                "code": "class Foo(*one, *two, *three): pass\n",
            },
            # Dictionary expansion render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    keywords=(cst.Arg(star="**", value=cst.Name("one")),),
                ),
                "code": "class Foo(**one): pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    keywords=(
                        cst.Arg(star="**", value=cst.Name("one")),
                        cst.Arg(star="**", value=cst.Name("two")),
                        cst.Arg(star="**", value=cst.Name("three")),
                    ),
                ),
                "code": "class Foo(**one, **two, **three): pass\n",
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    type_parameters=cst.TypeParameters(
                        (
                            cst.TypeParam(
                                cst.TypeVar(
                                    cst.Name("T"),
                                    bound=cst.Name("int"),
                                    colon=cst.Colon(
                                        whitespace_after=cst.SimpleWhitespace(" ")
                                    ),
                                ),
                                cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                            ),
                            cst.TypeParam(
                                cst.TypeVarTuple(cst.Name("Ts")),
                                cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                            ),
                            cst.TypeParam(cst.ParamSpec(cst.Name("KW"))),
                        )
                    ),
                ),
                "code": "class Foo[T: int, *Ts, **KW]: pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    type_parameters=cst.TypeParameters(
                        params=(
                            cst.TypeParam(
                                param=cst.TypeVar(
                                    cst.Name("T"),
                                    bound=cst.Name("str"),
                                    colon=cst.Colon(
                                        whitespace_before=cst.SimpleWhitespace(" "),
                                        whitespace_after=cst.ParenthesizedWhitespace(
                                            empty_lines=(cst.EmptyLine(),),
                                            indent=True,
                                        ),
                                    ),
                                ),
                                comma=cst.Comma(cst.SimpleWhitespace(" ")),
                            ),
                            cst.TypeParam(
                                cst.ParamSpec(
                                    cst.Name("PS"), cst.SimpleWhitespace("  ")
                                ),
                                cst.Comma(cst.SimpleWhitespace("  ")),
                            ),
                        )
                    ),
                    whitespace_after_type_parameters=cst.SimpleWhitespace("  "),
                ),
                "code": "class Foo[T :\n\nstr ,**  PS  ,]  : pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    type_parameters=cst.TypeParameters(
                        params=(
                            cst.TypeParam(
                                param=cst.TypeVar(
                                    cst.Name("T"),
                                    bound=cst.Name("str"),
                                    colon=cst.Colon(
                                        whitespace_before=cst.SimpleWhitespace(" "),
                                        whitespace_after=cst.ParenthesizedWhitespace(
                                            empty_lines=(cst.EmptyLine(),),
                                            indent=True,
                                        ),
                                    ),
                                ),
                                comma=cst.Comma(cst.SimpleWhitespace(" ")),
                            ),
                            cst.TypeParam(
                                cst.ParamSpec(
                                    cst.Name("PS"), cst.SimpleWhitespace("  ")
                                ),
                                cst.Comma(cst.SimpleWhitespace("  ")),
                            ),
                        )
                    ),
                    lpar=cst.LeftParen(),
                    rpar=cst.RightParen(),
                    whitespace_after_type_parameters=cst.SimpleWhitespace("  "),
                ),
                "code": "class Foo[T :\n\nstr ,**  PS  ,]  (): pass\n",
            },
        )
    )
    def test_valid_native(self, **kwargs: Any) -> None:
        if not is_native():
            self.skipTest("Disabled for pure python parser")
        self.validate_node(**kwargs)

    @data_provider(
        (
            # Basic parenthesis tests.
            (
                lambda: cst.ClassDef(
                    name=cst.Name("Foo"),
                    body=cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                ),
                "Do not mix concrete LeftParen/RightParen with MaybeSentinel",
            ),
            (
                lambda: cst.ClassDef(
                    name=cst.Name("Foo"),
                    body=cst.SimpleStatementSuite((cst.Pass(),)),
                    rpar=cst.RightParen(),
                ),
                "Do not mix concrete LeftParen/RightParen with MaybeSentinel",
            ),
            # Whitespace validation
            (
                lambda: cst.ClassDef(
                    name=cst.Name("Foo"),
                    body=cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_after_class=cst.SimpleWhitespace(""),
                ),
                "at least one space between 'class' and name",
            ),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)


class ClassDefParserTest(CSTNodeTest):
    @data_provider(
        (
            # Simple classdef
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"), cst.SimpleStatementSuite((cst.Pass(),))
                ),
                "code": "class Foo: pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(): pass\n",
            },
            # Positional arguments render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    bases=(cst.Arg(cst.Name("obj")),),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(obj): pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    bases=(
                        cst.Arg(
                            cst.Name("Bar"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            cst.Name("Baz"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(cst.Name("object")),
                    ),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(Bar, Baz, object): pass\n",
            },
            # Keyword arguments render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    keywords=(
                        cst.Arg(
                            keyword=cst.Name("metaclass"),
                            equal=cst.AssignEqual(),
                            value=cst.Name("Bar"),
                        ),
                    ),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(metaclass = Bar): pass\n",
            },
            # Iterator expansion render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    bases=(cst.Arg(star="*", value=cst.Name("one")),),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(*one): pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    bases=(
                        cst.Arg(
                            star="*",
                            value=cst.Name("one"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="*",
                            value=cst.Name("two"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(star="*", value=cst.Name("three")),
                    ),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(*one, *two, *three): pass\n",
            },
            # Dictionary expansion render test
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    keywords=(cst.Arg(star="**", value=cst.Name("one")),),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(**one): pass\n",
            },
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    keywords=(
                        cst.Arg(
                            star="**",
                            value=cst.Name("one"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(
                            star="**",
                            value=cst.Name("two"),
                            comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
                        ),
                        cst.Arg(star="**", value=cst.Name("three")),
                    ),
                    rpar=cst.RightParen(),
                ),
                "code": "class Foo(**one, **two, **three): pass\n",
            },
            # Decorator render tests
            {
                "node": cst.ClassDef(
                    cst.Name("Foo"),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    decorators=(cst.Decorator(cst.Name("foo")),),
                    lpar=cst.LeftParen(),
                    rpar=cst.RightParen(),
                ),
                "code": "@foo\nclass Foo(): pass\n",
                "expected_position": CodeRange((2, 0), (2, 17)),
            },
            {
                "node": cst.ClassDef(
                    leading_lines=(
                        cst.EmptyLine(),
                        cst.EmptyLine(comment=cst.Comment("# leading comment 1")),
                    ),
                    decorators=(
                        cst.Decorator(cst.Name("foo"), leading_lines=()),
                        cst.Decorator(
                            cst.Name("bar"),
                            leading_lines=(
                                cst.EmptyLine(
                                    comment=cst.Comment("# leading comment 2")
                                ),
                            ),
                        ),
                        cst.Decorator(
                            cst.Name("baz"),
                            leading_lines=(
                                cst.EmptyLine(
                                    comment=cst.Comment("# leading comment 3")
                                ),
                            ),
                        ),
                    ),
                    lines_after_decorators=(
                        cst.EmptyLine(comment=cst.Comment("# class comment")),
                    ),
                    name=cst.Name("Foo"),
                    body=cst.SimpleStatementSuite((cst.Pass(),)),
                    lpar=cst.LeftParen(),
                    rpar=cst.RightParen(),
                ),
                "code": "\n# leading comment 1\n@foo\n# leading comment 2\n@bar\n# leading comment 3\n@baz\n# class comment\nclass Foo(): pass\n",
                "expected_position": CodeRange((9, 0), (9, 17)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs, parser=parse_statement)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import cast, Tuple

import libcst as cst
from libcst import parse_module, parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst._parser.entrypoints import is_native
from libcst.metadata import CodeRange, MetadataWrapper, PositionProvider
from libcst.testing.utils import data_provider


class ModuleTest(CSTNodeTest):
    @data_provider(
        (
            # simplest possible program
            (cst.Module((cst.SimpleStatementLine((cst.Pass(),)),)), "pass\n"),
            # test default_newline
            (
                cst.Module(
                    (cst.SimpleStatementLine((cst.Pass(),)),), default_newline="\r"
                ),
                "pass\r",
            ),
            # test header/footer
            (
                cst.Module(
                    (cst.SimpleStatementLine((cst.Pass(),)),),
                    header=(cst.EmptyLine(comment=cst.Comment("# header")),),
                    footer=(cst.EmptyLine(comment=cst.Comment("# footer")),),
                ),
                "# header\npass\n# footer\n",
            ),
            # test has_trailing_newline
            (
                cst.Module(
                    (cst.SimpleStatementLine((cst.Pass(),)),),
                    has_trailing_newline=False,
                ),
                "pass",
            ),
            # an empty file
            (cst.Module((), has_trailing_newline=False), ""),
            # a file with only comments
            (
                cst.Module(
                    (),
                    header=(
                        cst.EmptyLine(comment=cst.Comment("# nothing to see here")),
                    ),
                ),
                "# nothing to see here\n",
            ),
            # TODO: test default_indent
        )
    )
    def test_code_and_bytes_properties(self, module: cst.Module, expected: str) -> None:
        self.assertEqual(module.code, expected)
        self.assertEqual(module.bytes, expected.encode("utf-8"))

    @data_provider(
        (
            (cst.Module(()), cst.Newline(), "\n"),
            (cst.Module((), default_newline="\r\n"), cst.Newline(), "\r\n"),
            # has_trailing_newline has no effect on code_for_node
            (cst.Module((), has_trailing_newline=False), cst.Newline(), "\n"),
            # TODO: test default_indent
        )
    )
    def test_code_for_node(
        self, module: cst.Module, node: cst.CSTNode, expected: str
    ) -> None:
        self.assertEqual(module.code_for_node(node), expected)

    @data_provider(
        {
            "empty_program": {
                "code": "",
                "expected": cst.Module([], has_trailing_newline=False),
            },
            "empty_program_with_newline": {
                "code": "\n",
                "expected": cst.Module([], has_trailing_newline=True),
                "enabled_for_native": False,
            },
            "empty_program_with_comments": {
                "code": "# some comment\n",
                "expected": cst.Module(
                    [], header=[cst.EmptyLine(comment=cst.Comment("# some comment"))]
                ),
            },
            "simple_pass": {
                "code": "pass\n",
                "expected": cst.Module([cst.SimpleStatementLine([cst.Pass()])]),
            },
            "simple_pass_with_header_footer": {
                "code": "# header\npass # trailing\n# footer\n",
                "expected": cst.Module(
                    [
                        cst.SimpleStatementLine(
                            [cst.Pass()],
                            trailing_whitespace=cst.TrailingWhitespace(
                                whitespace=cst.SimpleWhitespace(" "),
                                comment=cst.Comment("# trailing"),
                            ),
                        )
                    ],
                    header=[cst.EmptyLine(comment=cst.Comment("# header"))],
                    footer=[cst.EmptyLine(comment=cst.Comment("# footer"))],
                ),
            },
        }
    )
    def test_parser(
        self, *, code: str, expected: cst.Module, enabled_for_native: bool = True
    ) -> None:
        if is_native() and not enabled_for_native:
            self.skipTest("Disabled for native parser")
        self.assertEqual(parse_module(code), expected)

    @data_provider(
        {
            "empty": {"code": "", "expected": CodeRange((1, 0), (1, 0))},
            "empty_with_newline": {"code": "\n", "expected": CodeRange((1, 0), (2, 0))},
            "empty_program_with_comments": {
                "code": "# 2345",
                "expected": CodeRange((1, 0), (2, 0)),
            },
            "simple_pass": {"code": "pass\n", "expected": CodeRange((1, 0), (2, 0))},
            "simple_pass_with_header_footer": {
                "code": "# header\npass # trailing\n# footer\n",
                "expected": CodeRange((1, 0), (4, 0)),
            },
        }
    )
    def test_module_position(self, *, code: str, expected: CodeRange) -> None:
        wrapper = MetadataWrapper(parse_module(code))
        positions = wrapper.resolve(PositionProvider)

        self.assertEqual(positions[wrapper.module], expected)

    def cmp_position(
        self, actual: CodeRange, start: Tuple[int, int], end: Tuple[int, int]
    ) -> None:
        self.assertEqual(actual, CodeRange(start, end))

    def test_function_position(self) -> None:
        wrapper = MetadataWrapper(parse_module("def foo():\n    pass"))
        module = wrapper.module
        positions = wrapper.resolve(PositionProvider)

        fn = cast(cst.FunctionDef, module.body[0])
        stmt = cast(cst.SimpleStatementLine, fn.body.body[0])
        pass_stmt = cast(cst.Pass, stmt.body[0])
        self.cmp_position(positions[stmt], (2, 4), (2, 8))
        self.cmp_position(positions[pass_stmt], (2, 4), (2, 8))

    def test_nested_indent_position(self) -> None:
        wrapper = MetadataWrapper(
            parse_module("if True:\n    if False:\n        x = 1\nelse:\n    return")
        )
        module = wrapper.module
        positions = wrapper.resolve(PositionProvider)

        outer_if = cast(cst.If, module.body[0])
        inner_if = cast(cst.If, outer_if.body.body[0])
        assign = cast(cst.SimpleStatementLine, inner_if.body.body[0]).body[0]

        outer_else = cast(cst.Else, outer_if.orelse)
        return_stmt = cast(cst.SimpleStatementLine, outer_else.body.body[0]).body[0]

        self.cmp_position(positions[outer_if], (1, 0), (5, 10))
        self.cmp_position(positions[inner_if], (2, 4), (3, 13))
        self.cmp_position(positions[assign], (3, 8), (3, 13))
        self.cmp_position(positions[outer_else], (4, 0), (5, 10))
        self.cmp_position(positions[return_stmt], (5, 4), (5, 10))

    def test_multiline_string_position(self) -> None:
        wrapper = MetadataWrapper(parse_module('"abc"\\\n"def"'))
        module = wrapper.module
        positions = wrapper.resolve(PositionProvider)

        stmt = cast(cst.SimpleStatementLine, module.body[0])
        expr = cast(cst.Expr, stmt.body[0])
        string = expr.value

        self.cmp_position(positions[stmt], (1, 0), (2, 5))
        self.cmp_position(positions[expr], (1, 0), (2, 5))
        self.cmp_position(positions[string], (1, 0), (2, 5))

    def test_module_config_for_parsing(self) -> None:
        module = parse_module("pass\r")
        statement = parse_statement(
            "if True:\r    pass", config=module.config_for_parsing
        )
        self.assertEqual(
            statement,
            cst.If(
                test=cst.Name(value="True"),
                body=cst.IndentedBlock(
                    body=[cst.SimpleStatementLine(body=[cst.Pass()])],
                    header=cst.TrailingWhitespace(
                        newline=cst.Newline(
                            # This would be "\r" if we didn't pass the module config forward.
                            value=None
                        )
                    ),
                ),
            ),
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
from tempfile import TemporaryDirectory

from libcst.helpers.paths import chdir
from libcst.testing.utils import UnitTest


class PathsTest(UnitTest):
    def test_chdir(self) -> None:
        with TemporaryDirectory() as td:
            tdp = Path(td).resolve()
            inner = tdp / "foo" / "bar"
            inner.mkdir(parents=True)

            with self.subTest("string paths"):
                cwd1 = Path.cwd()

                with chdir(tdp.as_posix()) as path2:
                    cwd2 = Path.cwd()
                    self.assertEqual(tdp, cwd2)
                    self.assertEqual(tdp, path2)

                    with chdir(inner.as_posix()) as path3:
                        cwd3 = Path.cwd()
                        self.assertEqual(inner, cwd3)
                        self.assertEqual(inner, path3)

                    cwd4 = Path.cwd()
                    self.assertEqual(tdp, cwd4)
                    self.assertEqual(cwd2, cwd4)

                cwd5 = Path.cwd()
                self.assertEqual(cwd1, cwd5)

            with self.subTest("pathlib objects"):
                cwd1 = Path.cwd()

                with chdir(tdp) as path2:
                    cwd2 = Path.cwd()
                    self.assertEqual(tdp, cwd2)
                    self.assertEqual(tdp, path2)

                    with chdir(inner) as path3:
                        cwd3 = Path.cwd()
                        self.assertEqual(inner, cwd3)
                        self.assertEqual(inner, path3)

                    cwd4 = Path.cwd()
                    self.assertEqual(tdp, cwd4)
                    self.assertEqual(cwd2, cwd4)

                cwd5 = Path.cwd()
                self.assertEqual(cwd1, cwd5)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

from typing import Sequence

from libcst._exceptions import ParserSyntaxError
from libcst._parser.parso.python.token import PythonTokenTypes
from libcst._parser.parso.utils import parse_version_string, PythonVersionInfo
from libcst._parser.types.whitespace_state import WhitespaceState
from libcst._parser.wrapped_tokenize import Token, tokenize
from libcst.testing.utils import data_provider, UnitTest

_PY38 = parse_version_string("3.8.0")
_PY37 = parse_version_string("3.7.0")
_PY36 = parse_version_string("3.6.0")
_PY35 = parse_version_string("3.5.0")


class WrappedTokenizeTest(UnitTest):
    maxDiff = 10000

    @data_provider(
        {
            "simple_py35": (
                "pass;\n",
                _PY35,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="pass",
                        start_pos=(1, 0),
                        end_pos=(1, 4),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=";",
                        start_pos=(1, 4),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 5),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(2, 0),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "with_indent_py35": (
                "if foo:\n    bar\n",
                _PY35,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="if",
                        start_pos=(1, 0),
                        end_pos=(1, 2),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 3),
                        end_pos=(1, 6),
                        whitespace_before=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 6),
                        end_pos=(1, 7),
                        whitespace_before=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 7),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 4),
                        end_pos=(2, 7),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 7),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "async_py35": (
                "async def foo():\n    return await bar\n",
                _PY35,
                (
                    Token(
                        type=PythonTokenTypes.ASYNC,
                        string="async",
                        start_pos=(1, 0),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="def",
                        start_pos=(1, 6),
                        end_pos=(1, 9),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 10),
                        end_pos=(1, 13),
                        whitespace_before=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string="(",
                        start_pos=(1, 13),
                        end_pos=(1, 14),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=")",
                        start_pos=(1, 14),
                        end_pos=(1, 15),
                        whitespace_before=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 15),
                        end_pos=(1, 16),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 16),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="return",
                        start_pos=(2, 4),
                        end_pos=(2, 10),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.AWAIT,
                        string="await",
                        start_pos=(2, 11),
                        end_pos=(2, 16),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 17),
                        end_pos=(2, 20),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 20),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "async_no_token_35": (
                "async;\n",
                _PY35,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="async",
                        start_pos=(1, 0),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=";",
                        start_pos=(1, 5),
                        end_pos=(1, 6),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 6),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(2, 0),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "simple_py36": (
                "pass;\n",
                _PY36,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="pass",
                        start_pos=(1, 0),
                        end_pos=(1, 4),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=";",
                        start_pos=(1, 4),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 5),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(2, 0),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "with_indent_py36": (
                "if foo:\n    bar\n",
                _PY36,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="if",
                        start_pos=(1, 0),
                        end_pos=(1, 2),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 3),
                        end_pos=(1, 6),
                        whitespace_before=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 6),
                        end_pos=(1, 7),
                        whitespace_before=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 7),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 4),
                        end_pos=(2, 7),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 7),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "async_py36": (
                "async def foo():\n    return await bar\n",
                _PY36,
                (
                    Token(
                        type=PythonTokenTypes.ASYNC,
                        string="async",
                        start_pos=(1, 0),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="def",
                        start_pos=(1, 6),
                        end_pos=(1, 9),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 10),
                        end_pos=(1, 13),
                        whitespace_before=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string="(",
                        start_pos=(1, 13),
                        end_pos=(1, 14),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=")",
                        start_pos=(1, 14),
                        end_pos=(1, 15),
                        whitespace_before=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 15),
                        end_pos=(1, 16),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 16),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="return",
                        start_pos=(2, 4),
                        end_pos=(2, 10),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.AWAIT,
                        string="await",
                        start_pos=(2, 11),
                        end_pos=(2, 16),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 17),
                        end_pos=(2, 20),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 20),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "async_no_token_36": (
                "async;\n",
                _PY36,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="async",
                        start_pos=(1, 0),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=";",
                        start_pos=(1, 5),
                        end_pos=(1, 6),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 6),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(2, 0),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "simple_py37": (
                "pass;\n",
                _PY37,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="pass",
                        start_pos=(1, 0),
                        end_pos=(1, 4),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=";",
                        start_pos=(1, 4),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 5),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(2, 0),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "with_indent_py37": (
                "if foo:\n    bar\n",
                _PY37,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="if",
                        start_pos=(1, 0),
                        end_pos=(1, 2),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 3),
                        end_pos=(1, 6),
                        whitespace_before=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 6),
                        end_pos=(1, 7),
                        whitespace_before=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 7),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 4),
                        end_pos=(2, 7),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 7),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "async_py37": (
                "async def foo():\n    return await bar\n",
                _PY37,
                (
                    Token(
                        type=PythonTokenTypes.ASYNC,
                        string="async",
                        start_pos=(1, 0),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="def",
                        start_pos=(1, 6),
                        end_pos=(1, 9),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 10),
                        end_pos=(1, 13),
                        whitespace_before=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string="(",
                        start_pos=(1, 13),
                        end_pos=(1, 14),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=")",
                        start_pos=(1, 14),
                        end_pos=(1, 15),
                        whitespace_before=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 15),
                        end_pos=(1, 16),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 16),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="return",
                        start_pos=(2, 4),
                        end_pos=(2, 10),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.AWAIT,
                        string="await",
                        start_pos=(2, 11),
                        end_pos=(2, 16),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 17),
                        end_pos=(2, 20),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 20),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "simple_py38": (
                "pass;\n",
                _PY38,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="pass",
                        start_pos=(1, 0),
                        end_pos=(1, 4),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=";",
                        start_pos=(1, 4),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=4, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 5),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(2, 0),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "with_indent_py38": (
                "if foo:\n    bar\n",
                _PY38,
                (
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="if",
                        start_pos=(1, 0),
                        end_pos=(1, 2),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 3),
                        end_pos=(1, 6),
                        whitespace_before=WhitespaceState(
                            line=1, column=2, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 6),
                        end_pos=(1, 7),
                        whitespace_before=WhitespaceState(
                            line=1, column=6, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 7),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1, column=7, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 4),
                        end_pos=(2, 7),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 7),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=7,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
            "async_py38": (
                "async def foo():\n    return await bar\n",
                _PY38,
                (
                    Token(
                        type=PythonTokenTypes.ASYNC,
                        string="async",
                        start_pos=(1, 0),
                        end_pos=(1, 5),
                        whitespace_before=WhitespaceState(
                            line=1, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="def",
                        start_pos=(1, 6),
                        end_pos=(1, 9),
                        whitespace_before=WhitespaceState(
                            line=1, column=5, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="foo",
                        start_pos=(1, 10),
                        end_pos=(1, 13),
                        whitespace_before=WhitespaceState(
                            line=1, column=9, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string="(",
                        start_pos=(1, 13),
                        end_pos=(1, 14),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=13,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=")",
                        start_pos=(1, 14),
                        end_pos=(1, 15),
                        whitespace_before=WhitespaceState(
                            line=1, column=14, absolute_indent="", is_parenthesized=True
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.OP,
                        string=":",
                        start_pos=(1, 15),
                        end_pos=(1, 16),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=15,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(1, 16),
                        end_pos=(2, 0),
                        whitespace_before=WhitespaceState(
                            line=1,
                            column=16,
                            absolute_indent="",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.INDENT,
                        string="",
                        start_pos=(2, 4),
                        end_pos=(2, 4),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent="    ",
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="return",
                        start_pos=(2, 4),
                        end_pos=(2, 10),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=0,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.AWAIT,
                        string="await",
                        start_pos=(2, 11),
                        end_pos=(2, 16),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=10,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NAME,
                        string="bar",
                        start_pos=(2, 17),
                        end_pos=(2, 20),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=16,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.NEWLINE,
                        string="\n",
                        start_pos=(2, 20),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=2,
                            column=20,
                            absolute_indent="    ",
                            is_parenthesized=False,
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.DEDENT,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                    Token(
                        type=PythonTokenTypes.ENDMARKER,
                        string="",
                        start_pos=(3, 0),
                        end_pos=(3, 0),
                        whitespace_before=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        whitespace_after=WhitespaceState(
                            line=3, column=0, absolute_indent="", is_parenthesized=False
                        ),
                        relative_indent=None,
                    ),
                ),
            ),
        }
    )
    def test_tokenize(
        self, code: str, ver: PythonVersionInfo, expected: Sequence[Token]
    ) -> None:
        tokens = tuple(tokenize(code, ver))
        self.assertSequenceEqual(tokens, expected)
        for a, b in zip(tokens, tokens[1:]):
            # These must be the same object, so if whitespace gets consumed (mutated) at
            # the end of token a, it shows up at the beginning of token b.
            self.assertIs(a.whitespace_after, b.whitespace_before)

    def test_errortoken(self) -> None:
        for version in [_PY36, _PY37, _PY38]:
            with self.assertRaisesRegex(ParserSyntaxError, "not a valid token"):
                # use tuple() to read everything
                # The copyright symbol isn't a valid token
                tuple(tokenize("\u00a9", version))

    def test_error_dedent(self) -> None:
        for version in [_PY36, _PY37, _PY38]:
            with self.assertRaisesRegex(ParserSyntaxError, "Inconsistent indentation"):
                # create some inconsistent indents to generate an ERROR_DEDENT token
                tuple(tokenize("    a\n  b", version))



# with_wickedness

with foo :
    pass

with foo, bar:
    pass

with (foo, bar):
    pass

with (foo, bar,):
    pass

with foo, bar as bar:
    pass

with (foo, bar as bar):
    pass

with (foo, bar as bar,):
    pass

async def f():
    async with foo:

        with bar:
            pass

    async with foo :
        pass

    async with foo, bar:
        pass

    async with (foo, bar):
        pass

    async with (foo, bar,):
        pass

    async with foo, bar as bar:
        pass

    async with (foo, bar as bar):
        pass

    async with (foo, bar as bar,):
        pass

    async  with  foo(1+1)  as  bar  , 1  as (a, b, ) , 2 as [a, b] , 3 as a[b]  :
        pass



import contextlib
import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Generator
from unittest import TestCase

from libcst import BaseExpression, Call, matchers as m, Name
from libcst.codemod import (
    CodemodContext,
    gather_files,
    parallel_exec_transform_with_prettyprint,
    VisitorBasedCodemodCommand,
)
from libcst.codemod.visitors import AddImportsVisitor


class PrintToPPrintCommand(VisitorBasedCodemodCommand):
    def leave_Call(self, original_node: Call, updated_node: Call) -> BaseExpression:
        if m.matches(updated_node, m.Call(func=m.Name("print"))):
            AddImportsVisitor.add_needed_import(
                self.context,
                "pprint",
                "pprint",
            )
            return updated_node.with_changes(func=Name("pprint"))
        return super().leave_Call(original_node, updated_node)


@contextlib.contextmanager
def temp_workspace() -> Generator[Path, None, None]:
    cwd = os.getcwd()
    with TemporaryDirectory() as temp_dir:
        try:
            ws = Path(temp_dir).resolve()
            os.chdir(ws)
            yield ws
        finally:
            os.chdir(cwd)


class ToolE2ETest(TestCase):
    def test_leaky_codemod(self) -> None:
        with temp_workspace() as tmp:
            # File to trigger codemod
            example: Path = tmp / "example.py"
            example.write_text("""print("Hello")""")
            # File that should not be modified
            other = tmp / "other.py"
            other.touch()
            # Just a dir named "dir.py", should be ignored
            adir = tmp / "dir.py"
            adir.mkdir()

            # Run command
            command_instance = PrintToPPrintCommand(CodemodContext())
            files = gather_files(".")
            result = parallel_exec_transform_with_prettyprint(
                command_instance,
                files,
                format_code=False,
                hide_progress=True,
            )

            print(result)

            # Check results
            self.assertEqual(2, result.successes)
            self.assertEqual(0, result.skips)
            self.assertEqual(0, result.failures)
            # Expect example.py to be modified
            self.assertIn(
                "from pprint import pprint",
                example.read_text(),
                "import missing in example.py",
            )
            # Expect other.py to NOT be modified
            self.assertNotIn(
                "from pprint import pprint",
                other.read_text(),
                "import found in other.py",
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from textwrap import dedent
from typing import Optional, Sequence, Type

from libcst import parse_module, PartialParserConfig
from libcst.codemod._codemod import Codemod
from libcst.codemod._context import CodemodContext
from libcst.codemod._runner import SkipFile
from libcst.testing.utils import UnitTest


# pyre-fixme[13]: This should be an ABC but there are metaclass conflicts due to
# the way we implement the data_provider decorator, so pyre complains about the
# uninitialized TRANSFORM below.
class _CodemodTest:
    """
    Mixin that can be added to a unit test framework in order to provide
    convenience features. This is provided as an internal-only feature so
    that CodemodTest can be used with other frameworks. This is necessary
    since we set a metaclass on our UnitTest implementation.
    """

    TRANSFORM: Type[Codemod] = ...

    @staticmethod
    def make_fixture_data(data: str) -> str:
        """
        Given a code string originting from a multi-line triple-quoted string,
        normalize the code using ``dedent`` and ensuring a trailing newline
        is present.
        """

        lines = dedent(data).split("\n")

        def filter_line(line: str) -> str:
            if len(line.strip()) == 0:
                return ""
            return line

        # Get rid of lines that are space only
        lines = [filter_line(line) for line in lines]

        # Get rid of leading and trailing newlines (because of """ style strings)
        while lines and lines[0] == "":
            lines = lines[1:]
        while lines and lines[-1] == "":
            lines = lines[:-1]

        code = "\n".join(lines)
        if not code.endswith("\n"):
            return code + "\n"
        else:
            return code

    def assertCodeEqual(self, expected: str, actual: str) -> None:
        """
        Given an expected and actual code string, makes sure they equal. This
        ensures that both the expected and actual are sanitized, so its safe to
        use this on strings that may have come from a triple-quoted multi-line
        string.
        """

        # pyre-ignore This mixin needs to be used with a UnitTest subclass.
        self.assertEqual(
            CodemodTest.make_fixture_data(expected),
            CodemodTest.make_fixture_data(actual),
        )

    def assertCodemod(
        self,
        before: str,
        after: str,
        *args: object,
        context_override: Optional[CodemodContext] = None,
        python_version: Optional[str] = None,
        expected_warnings: Optional[Sequence[str]] = None,
        expected_skip: bool = False,
        **kwargs: object,
    ) -> None:
        """
        Given a before and after code string, and any args/kwargs that should
        be passed to the codemod constructor specified in
        :attr:`~CodemodTest.TRANSFORM`, validate that the codemod executes as
        expected. Verify that the codemod completes successfully, unless the
        ``expected_skip`` option is set to ``True``, in which case verify that
        the codemod skips.  Optionally, a :class:`CodemodContext` can be provided.
        If none is specified, a default, empty context is created for you.
        Additionally, the python version for the code parser can be overridden
        to a valid python version string such as `"3.6"`. If none is specified,
        the version of the interpreter running your tests will be used. Also, a
        list of warning strings can be specified and :meth:`~CodemodTest.assertCodemod`
        will verify that the codemod generates those warnings in the order
        specified. If it is left out, warnings are not checked.
        """

        context = context_override if context_override is not None else CodemodContext()
        # pyre-fixme[45]: Cannot instantiate abstract class `Codemod`.
        transform_instance = self.TRANSFORM(context, *args, **kwargs)
        input_tree = parse_module(
            CodemodTest.make_fixture_data(before),
            config=(
                PartialParserConfig(python_version=python_version)
                if python_version is not None
                else PartialParserConfig()
            ),
        )
        try:
            output_tree = transform_instance.transform_module(input_tree)
        except SkipFile:
            if not expected_skip:
                raise
            output_tree = input_tree
        else:
            if expected_skip:
                # pyre-ignore This mixin needs to be used with a UnitTest subclass.
                self.fail("Expected SkipFile but was not raised")
        # pyre-ignore This mixin needs to be used with a UnitTest subclass.
        self.assertEqual(
            CodemodTest.make_fixture_data(after),
            CodemodTest.make_fixture_data(output_tree.code),
        )
        if expected_warnings is not None:
            # pyre-ignore This mixin needs to be used with a UnitTest subclass.
            self.assertSequenceEqual(expected_warnings, context.warnings)


class CodemodTest(_CodemodTest, UnitTest):
    """
    Base test class for a :class:`Codemod` test. Provides facilities for
    auto-instantiating and executing a codemod, given the args/kwargs that
    should be passed to it. Set the :attr:`~CodemodTest.TRANSFORM` class
    attribute to the :class:`Codemod` class you wish to test and call
    :meth:`~CodemodTest.assertCodemod` inside your test method to verify it
    transforms various source code chunks correctly.

    Note that this is a subclass of ``UnitTest`` so any :class:`CodemodTest`
    can be executed using your favorite test runner such as the ``unittest``
    module.
    """



class Foo: ...

class Bar  :
    ...

class Old (  )  :
    gold : int


class OO ( Foo ) : ...

class OOP ( Foo , Bar, ) : pass

class OOPS (
    Foo ,

) :
    pass

class OOPSI ( Foo, * Bar , metaclass =
    foo ,
): pass

class OOPSIE ( list , *args, kw = arg , ** kwargs ) :
    what : does_this_even = mean

    def __init__(self) -> None:
        self.foo: Bar = Bar()




# foo

match  (  foo )   :  #comment

# more comments
    case  False  :  # comment

      ...
    case  (  True )  : ...
    case _ : ...
    case ( _  ) :  ... # foo

# bar

match x:
    case  "StringMatchValue"  : pass
    case   [1, 2]   : pass
    case   [  1 ,   * foo  , *  _ , ]: pass
    case  [  [ _,  ] , *_ ]: pass
    case  {1: _, 2: _}: pass
    case {  "foo" :  bar ,  ** rest }  : pass
    case { 1 : {**rest} ,  } :  pass
    case Point2D(): pass
    case  Cls (  0 , ) : pass
    case Cls  ( x=0,  y =  2)  :pass
    case  Cls  (  0  ,  1  ,  x  =  0  ,  y  =  2  )  :  pass
    case [x] as y: pass
    case  [x]   as  y    :  pass
    case (True)as x:pass
    case Foo:pass
    case (Foo):pass
    case (  Foo   ) :  pass
    case [  ( Foo  )   , ]: pass
    case Foo|Bar|Baz : pass
    case Foo  |    Bar |  (  Baz):  pass
    case x,y  ,  * more   :pass
    case y.z: pass
    case 1, 2: pass




# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest, parse_expression_as
from libcst._parser.entrypoints import is_native
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class DictTest(CSTNodeTest):
    @data_provider(
        [
            # zero-element dict
            {
                "node": cst.Dict([]),
                "code": "{}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 2)),
            },
            # one-element dict, sentinel comma value
            {
                "node": cst.Dict([cst.DictElement(cst.Name("k"), cst.Name("v"))]),
                "code": "{k: v}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 6)),
            },
            {
                "node": cst.Dict([cst.StarredDictElement(cst.Name("expanded"))]),
                "code": "{**expanded}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 12)),
            },
            # two-element dict, sentinel comma value
            {
                "node": cst.Dict(
                    [
                        cst.DictElement(cst.Name("k1"), cst.Name("v1")),
                        cst.DictElement(cst.Name("k2"), cst.Name("v2")),
                    ]
                ),
                "code": "{k1: v1, k2: v2}",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 16)),
            },
            # custom whitespace between brackets
            {
                "node": cst.Dict(
                    [cst.DictElement(cst.Name("k"), cst.Name("v"))],
                    lbrace=cst.LeftCurlyBrace(
                        whitespace_after=cst.SimpleWhitespace("\t")
                    ),
                    rbrace=cst.RightCurlyBrace(
                        whitespace_before=cst.SimpleWhitespace("\t\t")
                    ),
                ),
                "code": "{\tk: v\t\t}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 9)),
            },
            # with parenthesis
            {
                "node": cst.Dict(
                    [cst.DictElement(cst.Name("k"), cst.Name("v"))],
                    lpar=[cst.LeftParen()],
                    rpar=[cst.RightParen()],
                ),
                "code": "({k: v})",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 7)),
            },
            # starred element
            {
                "node": cst.Dict(
                    [
                        cst.StarredDictElement(cst.Name("one")),
                        cst.StarredDictElement(cst.Name("two")),
                    ]
                ),
                "code": "{**one, **two}",
                "parser": None,
                "expected_position": CodeRange((1, 0), (1, 14)),
            },
            # custom comma on DictElement
            {
                "node": cst.Dict(
                    [cst.DictElement(cst.Name("k"), cst.Name("v"), comma=cst.Comma())]
                ),
                "code": "{k: v,}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 7)),
            },
            # custom comma on StarredDictElement
            {
                "node": cst.Dict(
                    [cst.StarredDictElement(cst.Name("expanded"), comma=cst.Comma())]
                ),
                "code": "{**expanded,}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 13)),
            },
            # custom whitespace on DictElement
            {
                "node": cst.Dict(
                    [
                        cst.DictElement(
                            cst.Name("k"),
                            cst.Name("v"),
                            whitespace_before_colon=cst.SimpleWhitespace("\t"),
                            whitespace_after_colon=cst.SimpleWhitespace("\t\t"),
                        )
                    ]
                ),
                "code": "{k\t:\t\tv}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 8)),
            },
            # custom whitespace on StarredDictElement
            {
                "node": cst.Dict(
                    [
                        cst.DictElement(
                            cst.Name("k"), cst.Name("v"), comma=cst.Comma()
                        ),
                        cst.StarredDictElement(
                            cst.Name("expanded"),
                            whitespace_before_value=cst.SimpleWhitespace("  "),
                        ),
                    ]
                ),
                "code": "{k: v,**  expanded}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 19)),
            },
            # missing spaces around dict is always okay
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        cst.Name("b"),
                        cst.Dict([cst.DictElement(cst.Name("k"), cst.Name("v"))]),
                        ifs=[
                            cst.CompIf(
                                cst.Name("c"),
                                whitespace_before=cst.SimpleWhitespace(""),
                            )
                        ],
                        whitespace_after_in=cst.SimpleWhitespace(""),
                    ),
                ),
                "parser": parse_expression,
                "code": "(a for b in{k: v}if c)",
            },
        ]
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        [
            # unbalanced Dict
            {
                "get_node": lambda: cst.Dict([], lpar=[cst.LeftParen()]),
                "expected_re": "left paren without right paren",
            }
        ]
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)

    @data_provider(
        (
            {
                "code": "{**{}}",
                "parser": parse_expression_as(python_version="3.5"),
                "expect_success": True,
            },
            {
                "code": "{**{}}",
                "parser": parse_expression_as(python_version="3.3"),
                "expect_success": False,
            },
        )
    )
    def test_versions(self, **kwargs: Any) -> None:
        if is_native() and not kwargs.get("expect_success", True):
            self.skipTest("parse errors are disabled for native parser")
        self.assert_parses(**kwargs)



# 'import' dotted_as_names
import sys
import time, sys
# 'from' dotted_name 'import' ('*' | '(' import_as_names ')' | import_as_names)
from time import time
from time import (time)
from sys import path, argv
from sys import (path, argv)
from sys import (path, argv,)
from sys import *


from a import (b, )
from . import a
from .a import b
from ... import a
from ...a import b
from .... import a
from ...... import a


# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

import textwrap
from contextlib import ExitStack
from types import MappingProxyType
from typing import (
    Any,
    cast,
    Collection,
    Iterable,
    Mapping,
    MutableMapping,
    MutableSet,
    Optional,
    Type,
    TYPE_CHECKING,
    TypeVar,
)

from libcst._batched_visitor import BatchableCSTVisitor, visit_batched, VisitorMethod
from libcst._exceptions import MetadataException
from libcst.metadata.base_provider import BatchableMetadataProvider

if TYPE_CHECKING:
    from libcst._nodes.base import CSTNode  # noqa: F401
    from libcst._nodes.module import Module  # noqa: F401
    from libcst._visitors import CSTVisitorT  # noqa: F401
    from libcst.metadata.base_provider import (  # noqa: F401
        BaseMetadataProvider,
        ProviderT,
    )


_T = TypeVar("_T")


def _gen_batchable(
    wrapper: "MetadataWrapper",
    # pyre-fixme[2]: Parameter `providers` must have a type that does not contain `Any`
    providers: Iterable[BatchableMetadataProvider[Any]],
) -> Mapping["ProviderT", Mapping["CSTNode", object]]:
    """
    Returns map of metadata mappings from resolving ``providers`` on ``wrapper``.
    """
    wrapper.visit_batched(providers)

    # Make immutable metadata mapping
    # pyre-ignore[7]
    return {type(p): MappingProxyType(dict(p._computed)) for p in providers}


def _gather_providers(
    providers: Collection["ProviderT"], gathered: MutableSet["ProviderT"]
) -> MutableSet["ProviderT"]:
    """
    Recursively gathers all the given providers and their dependencies.
    """
    for P in providers:
        if P not in gathered:
            gathered.add(P)
            _gather_providers(P.METADATA_DEPENDENCIES, gathered)
    return gathered


def _resolve_impl(
    wrapper: "MetadataWrapper", providers: Collection["ProviderT"]
) -> None:
    """
    Updates the _metadata map on wrapper with metadata from the given providers
    as well as their dependencies.
    """
    completed = set(wrapper._metadata.keys())
    remaining = _gather_providers(set(providers), set()) - completed

    while len(remaining) > 0:
        batchable = set()

        for P in remaining:
            if set(P.METADATA_DEPENDENCIES).issubset(completed):
                if issubclass(P, BatchableMetadataProvider):
                    batchable.add(P)
                else:
                    wrapper._metadata[P] = (
                        P(wrapper._cache.get(P))._gen(wrapper)
                        if P.gen_cache
                        else P()._gen(wrapper)
                    )
                    completed.add(P)

        initialized_batchable = [
            p(wrapper._cache.get(p)) if p.gen_cache else p() for p in batchable
        ]
        metadata_batch = _gen_batchable(wrapper, initialized_batchable)
        wrapper._metadata.update(metadata_batch)
        completed |= batchable

        if len(completed) == 0 and len(batchable) == 0:
            # remaining must be non-empty at this point
            names = ", ".join([P.__name__ for P in remaining])
            raise MetadataException(f"Detected circular dependencies in {names}")

        remaining -= completed


class MetadataWrapper:
    """
    A wrapper around a :class:`~libcst.Module` that stores associated metadata
    for that module.

    When a :class:`MetadataWrapper` is constructed over a module, the wrapper will
    store a deep copy of the original module. This means
    ``MetadataWrapper(module).module == module`` is ``False``.

    This copying operation ensures that a node will never appear twice (by identity) in
    the same tree. This allows us to uniquely look up metadata for a node based on a
    node's identity.
    """

    __slots__ = ["__module", "_metadata", "_cache"]

    __module: "Module"
    _metadata: MutableMapping["ProviderT", Mapping["CSTNode", object]]
    _cache: Mapping["ProviderT", object]

    def __init__(
        self,
        module: "Module",
        unsafe_skip_copy: bool = False,
        cache: Mapping["ProviderT", object] = {},
    ) -> None:
        """
        :param module: The module to wrap. This is deeply copied by default.
        :param unsafe_skip_copy: When true, this skips the deep cloning of the module.
            This can provide a small performance benefit, but you should only use this
            if you know that there are no duplicate nodes in your tree (e.g. this
            module came from the parser).
        :param cache: Pass the needed cache to wrapper to be used when resolving metadata.
        """
        # Ensure that module is safe to use by copying the module to remove
        # any duplicate nodes.
        if not unsafe_skip_copy:
            module = module.deep_clone()
        self.__module = module
        self._metadata = {}
        self._cache = cache

    def __repr__(self) -> str:
        return f"MetadataWrapper(\n{textwrap.indent(repr(self.module), ' ' * 4)},\n)"

    @property
    def module(self) -> "Module":
        """
        The module that's wrapped by this MetadataWrapper. By default, this is a deep
        copy of the passed in module.

        ::

            mw = ModuleWrapper(module)
            # Because `mw.module is not module`, you probably want to do visit and do
            # your analysis on `mw.module`, not `module`.
            mw.module.visit(DoSomeAnalysisVisitor)
        """
        # use a property getter to enforce that this is a read-only variable
        return self.__module

    def resolve(
        self, provider: Type["BaseMetadataProvider[_T]"]
    ) -> Mapping["CSTNode", _T]:
        """
        Returns a copy of the metadata mapping computed by ``provider``.
        """
        if provider in self._metadata:
            metadata = self._metadata[provider]
        else:
            metadata = self.resolve_many([provider])[provider]

        return cast(Mapping["CSTNode", _T], metadata)

    def resolve_many(
        self, providers: Collection["ProviderT"]
    ) -> Mapping["ProviderT", Mapping["CSTNode", object]]:
        """
        Returns a copy of the map of metadata mapping computed by each provider
        in ``providers``.

        The returned map does not contain any metadata from undeclared metadata
        dependencies that ``providers`` has.
        """
        _resolve_impl(self, providers)

        # Only return what what declared in providers
        return {k: self._metadata[k] for k in providers}

    def visit(self, visitor: "CSTVisitorT") -> "Module":
        """
        Convenience method to resolve metadata before performing a traversal over
        ``self.module`` with ``visitor``. See :func:`~libcst.Module.visit`.
        """
        with visitor.resolve(self):
            return self.module.visit(visitor)

    def visit_batched(
        self,
        visitors: Iterable[BatchableCSTVisitor],
        before_visit: Optional[VisitorMethod] = None,
        after_leave: Optional[VisitorMethod] = None,
    ) -> "CSTNode":
        """
        Convenience method to resolve metadata before performing a traversal over
        ``self.module`` with ``visitors``. See :func:`~libcst.visit_batched`.
        """
        with ExitStack() as stack:
            # Resolve dependencies of visitors
            for v in visitors:
                stack.enter_context(v.resolve(self))

            return visit_batched(self.module, visitors, before_visit, after_leave)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from dataclasses import dataclass, fields
from typing import Generator, List, Optional, Sequence, Set, Tuple, Type, Union

import libcst as cst
from libcst import ensure_type, parse_expression
from libcst.codegen.gather import all_libcst_nodes, typeclasses

CST_DIR: Set[str] = set(dir(cst))
CLASS_RE = r"<class \'(.*?)\'>"
OPTIONAL_RE = r"typing\.Union\[([^,]*?), NoneType]"


class CleanseFullTypeNames(cst.CSTTransformer):
    def leave_Call(
        self, original_node: cst.Call, updated_node: cst.Call
    ) -> cst.BaseExpression:
        # Convert forward ref repr back to a SimpleString.
        if isinstance(updated_node.func, cst.Name) and (
            updated_node.func.deep_equals(cst.Name("_ForwardRef"))
            or updated_node.func.deep_equals(cst.Name("ForwardRef"))
        ):
            return updated_node.args[0].value
        return updated_node

    def leave_Attribute(
        self, original_node: cst.Attribute, updated_node: cst.Attribute
    ) -> Union[cst.Attribute, cst.Name]:
        # Unwrap all attributes, so things like libcst.x.y.Name becomes Name
        return updated_node.attr

    def leave_Name(
        self, original_node: cst.Name, updated_node: cst.Name
    ) -> Union[cst.Name, cst.SimpleString]:
        value = updated_node.value
        if value == "NoneType":
            # This is special-cased in typing, un-special case it.
            return updated_node.with_changes(value="None")
        if value in CST_DIR and not value.endswith("Sentinel"):
            # If this isn't a typing define and it isn't a builtin, convert it to
            # a forward ref string.
            return cst.SimpleString(repr(value))
        return updated_node

    def leave_SubscriptElement(
        self, original_node: cst.SubscriptElement, updated_node: cst.SubscriptElement
    ) -> Union[cst.SubscriptElement, cst.RemovalSentinel]:
        slc = updated_node.slice
        if isinstance(slc, cst.Index):
            val = slc.value
            if isinstance(val, cst.Name):
                if "Sentinel" in val.value:
                    # We don't support maybes in matchers.
                    return cst.RemoveFromParent()
        # Simple trick to kill trailing commas
        return updated_node.with_changes(comma=cst.MaybeSentinel.DEFAULT)


class RemoveTypesFromGeneric(cst.CSTTransformer):
    def __init__(self, values: Sequence[str]) -> None:
        self.values: Set[str] = set(values)

    def leave_SubscriptElement(
        self, original_node: cst.SubscriptElement, updated_node: cst.SubscriptElement
    ) -> Union[cst.SubscriptElement, cst.RemovalSentinel]:
        slc = updated_node.slice
        if isinstance(slc, cst.Index):
            val = slc.value
            if isinstance(val, cst.Name):
                if val.value in self.values:
                    # This type matches, so out it goes
                    return cst.RemoveFromParent()
        return updated_node


def _remove_types(
    oldtype: cst.BaseExpression, values: Sequence[str]
) -> cst.BaseExpression:
    """
    Given a BaseExpression from a type, return a new BaseExpression that does not
    refer to any types listed in values.
    """
    return ensure_type(
        oldtype.visit(RemoveTypesFromGeneric(values)), cst.BaseExpression
    )


class MatcherClassToLibCSTClass(cst.CSTTransformer):
    def leave_SimpleString(
        self, original_node: cst.SimpleString, updated_node: cst.SimpleString
    ) -> Union[cst.SimpleString, cst.Attribute]:
        value = updated_node.evaluated_value
        if value in CST_DIR:
            return cst.Attribute(cst.Name("cst"), cst.Name(value))
        return updated_node


def _convert_match_nodes_to_cst_nodes(
    matchtype: cst.BaseExpression,
) -> cst.BaseExpression:
    """
    Given a BaseExpression in a type, convert this to a new BaseExpression that refers
    to LibCST nodes instead of forward references to matcher nodes.
    """
    return ensure_type(matchtype.visit(MatcherClassToLibCSTClass()), cst.BaseExpression)


def _get_match_if_true(oldtype: cst.BaseExpression) -> cst.SubscriptElement:
    """
    Construct a MatchIfTrue type node appropriate for going into a Union.
    """
    return cst.SubscriptElement(
        cst.Index(
            cst.Subscript(
                cst.Name("MatchIfTrue"),
                slice=(
                    cst.SubscriptElement(
                        cst.Index(
                            # MatchIfTrue takes in the original node type,
                            # and returns a boolean. So, lets convert our
                            # quoted classes (forward refs to other
                            # matchers) back to the CSTNode they refer to.
                            # We can do this because there's always a 1:1
                            # name mapping.
                            _convert_match_nodes_to_cst_nodes(oldtype)
                        ),
                    ),
                ),
            )
        )
    )


def _add_generic(name: str, oldtype: cst.BaseExpression) -> cst.BaseExpression:
    return cst.Subscript(cst.Name(name), (cst.SubscriptElement(cst.Index(oldtype)),))


class AddLogicMatchersToUnions(cst.CSTTransformer):
    def leave_Subscript(
        self, original_node: cst.Subscript, updated_node: cst.Subscript
    ) -> cst.Subscript:
        if updated_node.value.deep_equals(cst.Name("Union")):
            # Take the original node, remove do not care so we have concrete types.
            # Explicitly taking the original node because we want to discard nested
            # changes.
            concrete_only_expr = _remove_types(updated_node, ["DoNotCareSentinel"])
            return updated_node.with_changes(
                slice=[
                    *updated_node.slice,
                    cst.SubscriptElement(
                        cst.Index(_add_generic("OneOf", concrete_only_expr))
                    ),
                    cst.SubscriptElement(
                        cst.Index(_add_generic("AllOf", concrete_only_expr))
                    ),
                ]
            )
        return updated_node


class AddWildcardsToSequenceUnions(cst.CSTTransformer):
    def __init__(self) -> None:
        super().__init__()
        self.in_match_if_true: Set[cst.CSTNode] = set()
        self.fixup_nodes: Set[cst.Subscript] = set()

    def visit_Subscript(self, node: cst.Subscript) -> None:
        # If the current node is a MatchIfTrue, we don't want to modify it.
        if node.value.deep_equals(cst.Name("MatchIfTrue")):
            self.in_match_if_true.add(node)
        # If the direct descendant is a union, lets add it to be fixed up.
        elif node.value.deep_equals(cst.Name("Sequence")):
            if self.in_match_if_true:
                # We don't want to add AtLeastN/AtMostN inside MatchIfTrue
                # type blocks, even for sequence types.
                return
            if len(node.slice) != 1:
                raise Exception(
                    "Unexpected number of sequence elements inside Sequence type "
                    + "annotation!"
                )
            nodeslice = node.slice[0].slice
            if isinstance(nodeslice, cst.Index):
                possibleunion = nodeslice.value
                if isinstance(possibleunion, cst.Subscript):
                    if possibleunion.value.deep_equals(cst.Name("Union")):
                        self.fixup_nodes.add(possibleunion)

    def leave_Subscript(
        self, original_node: cst.Subscript, updated_node: cst.Subscript
    ) -> cst.Subscript:
        if original_node in self.in_match_if_true:
            self.in_match_if_true.remove(original_node)
        if original_node in self.fixup_nodes:
            self.fixup_nodes.remove(original_node)
            return updated_node.with_changes(
                slice=[
                    *updated_node.slice,
                    cst.SubscriptElement(
                        cst.Index(_add_generic("AtLeastN", original_node))
                    ),
                    cst.SubscriptElement(
                        cst.Index(_add_generic("AtMostN", original_node))
                    ),
                ]
            )
        return updated_node


def _get_do_not_care() -> cst.SubscriptElement:
    """
    Construct a DoNotCareSentinel entry appropriate for going into a Union.
    """

    return cst.SubscriptElement(cst.Index(cst.Name("DoNotCareSentinel")))


def _get_match_metadata() -> cst.SubscriptElement:
    """
    Construct a MetadataMatchType entry appropriate for going into a Union.
    """

    return cst.SubscriptElement(cst.Index(cst.Name("MetadataMatchType")))


def _get_wrapped_union_type(
    node: cst.BaseExpression,
    addition: cst.SubscriptElement,
    *additions: cst.SubscriptElement,
) -> cst.Subscript:
    """
    Take two or more nodes, wrap them in a union type. Function signature is
    explicitly defined as taking at least one addition for type safety.

    """

    return cst.Subscript(
        cst.Name("Union"), [cst.SubscriptElement(cst.Index(node)), addition, *additions]
    )


# List of global aliases we've already generated, so we don't redefine types
_global_aliases: Set[str] = set()


@dataclass(frozen=True)
class Alias:
    name: str
    type: str


@dataclass(frozen=True)
class Field:
    name: str
    type: str
    aliases: List[Alias]


def _get_raw_name(node: cst.CSTNode) -> Optional[str]:
    if isinstance(node, cst.Name):
        return node.value
    elif isinstance(node, cst.SimpleString):
        evaluated_value = node.evaluated_value
        if isinstance(evaluated_value, str):
            return evaluated_value
    elif isinstance(node, cst.SubscriptElement):
        return _get_raw_name(node.slice)
    elif isinstance(node, cst.Index):
        return _get_raw_name(node.value)
    else:
        return None


def _get_alias_name(node: cst.CSTNode) -> Optional[str]:
    if isinstance(node, (cst.Name, cst.SimpleString)):
        return f"{_get_raw_name(node)}MatchType"
    elif isinstance(node, cst.Subscript):
        if node.value.deep_equals(cst.Name("Union")):
            names = [_get_raw_name(s) for s in node.slice]
            if any(n is None for n in names):
                return None
            return "Or".join(n for n in names if n is not None) + "MatchType"

    return None


def _wrap_clean_type(
    aliases: List[Alias], name: Optional[str], value: cst.Subscript
) -> cst.BaseExpression:
    if name is not None:
        # We created an alias, lets use that, wrapping the alias in a do not care.
        aliases.append(Alias(name=name, type=cst.Module(body=()).code_for_node(value)))
        return _get_wrapped_union_type(cst.Name(name), _get_do_not_care())
    else:
        # Couldn't name the alias, fall back to regular node creation, add do not
        # care to the resulting type we widened.
        return value.with_changes(slice=[*value.slice, _get_do_not_care()])


def _get_clean_type_from_expression(
    aliases: List[Alias], typecst: cst.BaseExpression
) -> cst.BaseExpression:
    name = _get_alias_name(typecst)
    value = _get_wrapped_union_type(
        typecst, _get_match_metadata(), _get_match_if_true(typecst)
    )
    return _wrap_clean_type(aliases, name, value)


def _maybe_fix_sequence_in_union(
    aliases: List[Alias], typecst: cst.SubscriptElement
) -> cst.SubscriptElement:
    slc = typecst.slice
    if isinstance(slc, cst.Index):
        val = slc.value
        if isinstance(val, cst.Subscript):
            return cst.ensure_type(
                typecst.deep_replace(val, _get_clean_type_from_subscript(aliases, val)),
                cst.SubscriptElement,
            )
    return typecst


def _get_clean_type_from_union(
    aliases: List[Alias], typecst: cst.Subscript
) -> cst.BaseExpression:
    name = _get_alias_name(typecst)
    value = typecst.with_changes(
        slice=[
            *[_maybe_fix_sequence_in_union(aliases, slc) for slc in typecst.slice],
            _get_match_metadata(),
            _get_match_if_true(typecst),
        ]
    )
    return _wrap_clean_type(aliases, name, value)


def _get_clean_type_from_subscript(
    aliases: List[Alias], typecst: cst.Subscript
) -> cst.BaseExpression:
    if typecst.value.deep_equals(cst.Name("Sequence")):
        # Lets attempt to widen the sequence type and alias it.
        if len(typecst.slice) != 1:
            raise Exception("Logic error, Sequence shouldn't have more than one param!")
        inner_type = typecst.slice[0].slice
        if not isinstance(inner_type, cst.Index):
            raise Exception("Logic error, expecting Index for only Sequence element!")
        inner_type = inner_type.value

        if isinstance(inner_type, cst.Subscript):
            clean_inner_type = _get_clean_type_from_subscript(aliases, inner_type)
        elif isinstance(inner_type, (cst.Name, cst.SimpleString)):
            clean_inner_type = _get_clean_type_from_expression(aliases, inner_type)
        else:
            raise Exception("Logic error, unexpected type in Sequence!")

        return _get_wrapped_union_type(
            typecst.deep_replace(inner_type, clean_inner_type),
            _get_do_not_care(),
            _get_match_if_true(typecst),
        )
    # We can modify this as-is to add our extra values
    elif typecst.value.deep_equals(cst.Name("Union")):
        return _get_clean_type_from_union(aliases, typecst)
    else:
        # Don't handle other types like "Literal", just widen them.
        return _get_clean_type_from_expression(aliases, typecst)


def _get_clean_type_and_aliases(
    typeobj: object,
) -> Tuple[str, List[Alias]]:  # noqa: C901
    """
    Given a type object as returned by dataclasses, sanitize it and convert it
    to a type string that is appropriate for our codegen below.
    """

    # First, get the type as a parseable expression.
    typestr = repr(typeobj)
    typestr = re.sub(CLASS_RE, r"\1", typestr)
    typestr = re.sub(OPTIONAL_RE, r"typing.Optional[\1]", typestr)

    # Now, parse the expression with LibCST.
    cleanser = CleanseFullTypeNames()
    typecst = parse_expression(typestr)
    typecst = typecst.visit(cleanser)
    aliases: List[Alias] = []

    # Now, convert the type to allow for MetadataMatchType and MatchIfTrue values.
    if isinstance(typecst, cst.Subscript):
        clean_type = _get_clean_type_from_subscript(aliases, typecst)
    elif isinstance(typecst, (cst.Name, cst.SimpleString)):
        clean_type = _get_clean_type_from_expression(aliases, typecst)
    else:
        raise Exception("Logic error, unexpected top level type!")

    # Now, insert OneOf/AllOf and MatchIfTrue into unions so we can typecheck their usage.
    # This allows us to put OneOf[SomeType] or MatchIfTrue[cst.SomeType] into any
    # spot that we would have originally allowed a SomeType.
    clean_type = ensure_type(clean_type.visit(AddLogicMatchersToUnions()), cst.CSTNode)
    # Now, insert AtMostN and AtLeastN into sequence unions, so we can typecheck
    # them. This relies on the previous OneOf/AllOf insertion to ensure that all
    # sequences we care about are Sequence[Union[<x>]].
    clean_type = ensure_type(
        clean_type.visit(AddWildcardsToSequenceUnions()), cst.CSTNode
    )
    # Finally, generate the code given a default Module so we can spit it out.
    return cst.Module(body=()).code_for_node(clean_type), aliases


def _get_fields(node: Type[cst.CSTNode]) -> Generator[Field, None, None]:
    """
    Given a CSTNode, generate a field name and type string for each.
    """

    for field in fields(node) or []:
        if field.name == "_metadata":
            continue

        fieldtype, aliases = _get_clean_type_and_aliases(field.type)
        yield Field(
            name=field.name,
            type=fieldtype,
            aliases=[a for a in aliases if a.name not in _global_aliases],
        )
        _global_aliases.update(a.name for a in aliases)


all_exports: Set[str] = set()
generated_code: List[str] = []
generated_code.append("# Copyright (c) Meta Platforms, Inc. and affiliates.")
generated_code.append("#")
generated_code.append(
    "# This source code is licensed under the MIT license found in the"
)
generated_code.append("# LICENSE file in the root directory of this source tree.")
generated_code.append("")
generated_code.append("")
generated_code.append("# This file was generated by libcst.codegen.gen_matcher_classes")
generated_code.append("from dataclasses import dataclass")
generated_code.append("from typing import Optional, Sequence, Union")
generated_code.append("from typing_extensions import Literal")
generated_code.append("import libcst as cst")
generated_code.append("")
generated_code.append(
    "from libcst.matchers._matcher_base import AbstractBaseMatcherNodeMeta, BaseMatcherNode, DoNotCareSentinel, DoNotCare, TypeOf, OneOf, AllOf, DoesNotMatch, MatchIfTrue, MatchRegex, MatchMetadata, MatchMetadataIfTrue, ZeroOrMore, AtLeastN, ZeroOrOne, AtMostN, SaveMatchedNode, extract, extractall, findall, matches, replace"
)
all_exports.update(
    [
        "BaseMatcherNode",
        "DoNotCareSentinel",
        "DoNotCare",
        "OneOf",
        "AllOf",
        "DoesNotMatch",
        "MatchIfTrue",
        "MatchRegex",
        "MatchMetadata",
        "MatchMetadataIfTrue",
        "TypeOf",
        "ZeroOrMore",
        "AtLeastN",
        "ZeroOrOne",
        "AtMostN",
        "SaveMatchedNode",
        "extract",
        "extractall",
        "findall",
        "matches",
        "replace",
    ]
)
generated_code.append(
    "from libcst.matchers._decorators import call_if_inside, call_if_not_inside, visit, leave"
)
all_exports.update(["call_if_inside", "call_if_not_inside", "visit", "leave"])
generated_code.append(
    "from libcst.matchers._visitors import MatchDecoratorMismatch, MatcherDecoratableTransformer, MatcherDecoratableVisitor"
)
all_exports.update(
    [
        "MatchDecoratorMismatch",
        "MatcherDecoratableTransformer",
        "MatcherDecoratableVisitor",
    ]
)

generated_code.append("")
generated_code.append("")
generated_code.append("class _NodeABC(metaclass=AbstractBaseMatcherNodeMeta):")
generated_code.append("    __slots__ = ()")

for base in typeclasses:
    generated_code.append("")
    generated_code.append("")
    generated_code.append(f"class {base.__name__}(_NodeABC):")
    generated_code.append("    pass")
    all_exports.add(base.__name__)


# Add a generic MetadataMatchType to be referred to by everywhere else.
generated_code.append("")
generated_code.append("")
generated_code.append("MetadataMatchType = Union[MatchMetadata, MatchMetadataIfTrue]")


for node in all_libcst_nodes:
    if node.__name__.startswith("Base"):
        continue
    classes: List[str] = []
    for tc in typeclasses:
        if issubclass(node, tc):
            classes.append(tc.__name__)
    classes.append("BaseMatcherNode")

    has_aliases = False
    node_fields = list(_get_fields(node))
    for field in node_fields:
        for alias in field.aliases:
            # Output a separator if we're going to output any aliases
            if not has_aliases:
                generated_code.append("")
                generated_code.append("")
                has_aliases = True

            # Must generate code for aliases before the class they are referenced in
            generated_code.append(f"{alias.name} = {alias.type}")

    generated_code.append("")
    generated_code.append("")
    generated_code.append("@dataclass(frozen=True, eq=False, unsafe_hash=False)")
    generated_code.append(f'class {node.__name__}({", ".join(classes)}):')
    all_exports.add(node.__name__)

    fields_printed = False
    for field in node_fields:
        fields_printed = True
        generated_code.append(f"    {field.name}: {field.type} = DoNotCare()")

    # Add special metadata field
    generated_code.append(
        "    metadata: Union[MetadataMatchType, DoNotCareSentinel, OneOf[MetadataMatchType], AllOf[MetadataMatchType]] = DoNotCare()"
    )


# Make sure to add an __all__ for flake8 and compatibility with "from libcst.matchers import *"
generated_code.append(f"__all__ = {repr(sorted(all_exports))}")


if __name__ == "__main__":
    # Output the code
    print("\n".join(generated_code))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class ReturnCreateTest(CSTNodeTest):
    @data_provider(
        (
            {
                "node": cst.SimpleStatementLine([cst.Return()]),
                "code": "return\n",
                "expected_position": CodeRange((1, 0), (1, 6)),
            },
            {
                "node": cst.SimpleStatementLine([cst.Return(cst.Name("abc"))]),
                "code": "return abc\n",
                "expected_position": CodeRange((1, 0), (1, 10)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": lambda: cst.Return(
                    cst.Name("abc"), whitespace_after_return=cst.SimpleWhitespace("")
                ),
                "expected_re": "Must have at least one space after 'return'.",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)


class ReturnParseTest(CSTNodeTest):
    @data_provider(
        (
            {
                "node": cst.SimpleStatementLine(
                    [cst.Return(whitespace_after_return=cst.SimpleWhitespace(""))]
                ),
                "code": "return\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    [
                        cst.Return(
                            cst.Name("abc"),
                            whitespace_after_return=cst.SimpleWhitespace(" "),
                        )
                    ]
                ),
                "code": "return abc\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    [
                        cst.Return(
                            cst.Name("abc"),
                            whitespace_after_return=cst.SimpleWhitespace("   "),
                        )
                    ]
                ),
                "code": "return   abc\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    [
                        cst.Return(
                            cst.Name(
                                "abc", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                            ),
                            whitespace_after_return=cst.SimpleWhitespace(""),
                        )
                    ]
                ),
                "code": "return(abc)\n",
                "parser": parse_statement,
            },
            {
                "node": cst.SimpleStatementLine(
                    [
                        cst.Return(
                            cst.Name("abc"),
                            whitespace_after_return=cst.SimpleWhitespace(" "),
                            semicolon=cst.Semicolon(),
                        )
                    ]
                ),
                "code": "return abc;\n",
                "parser": parse_statement,
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
import itertools
import re
from typing import Callable, cast, List, Sequence

import libcst as cst
import libcst.matchers as m
from libcst.codemod import VisitorBasedCodemodCommand

USE_FSTRING_SIMPLE_EXPRESSION_MAX_LENGTH = 30


def _match_simple_string(node: cst.CSTNode) -> bool:
    if isinstance(node, cst.SimpleString) and not node.prefix.lower().startswith("b"):
        # SimpleString can be a bytes and fstring don't support bytes
        return re.fullmatch("[^%]*(%s[^%]*)+", node.raw_value) is not None
    return False


def _gen_match_simple_expression(module: cst.Module) -> Callable[[cst.CSTNode], bool]:
    def _match_simple_expression(node: cst.CSTNode) -> bool:
        # either each element in Tuple is simple expression or the entire expression is simple.
        if (
            isinstance(node, cst.Tuple)
            and all(
                len(module.code_for_node(elm.value))
                < USE_FSTRING_SIMPLE_EXPRESSION_MAX_LENGTH
                for elm in node.elements
            )
        ) or len(module.code_for_node(node)) < USE_FSTRING_SIMPLE_EXPRESSION_MAX_LENGTH:
            return True
        return False

    return _match_simple_expression


class EscapeStringQuote(cst.CSTTransformer):
    def __init__(self, quote: str) -> None:
        self.quote = quote
        super().__init__()

    def leave_SimpleString(
        self, original_node: cst.SimpleString, updated_node: cst.SimpleString
    ) -> cst.SimpleString:
        if self.quote == original_node.quote:
            for quo in ["'", '"', "'''", '"""']:
                if quo != original_node.quote and quo not in original_node.raw_value:
                    escaped_string = cst.SimpleString(
                        original_node.prefix + quo + original_node.raw_value + quo
                    )
                    if escaped_string.evaluated_value != original_node.evaluated_value:
                        raise Exception(
                            f"Failed to escape string:\n  original:{original_node.value}\n  escaped:{escaped_string.value}"
                        )
                    else:
                        return escaped_string
            raise Exception(
                f"Cannot find a good quote for escaping the SimpleString: {original_node.value}"
            )
        return original_node


class ConvertPercentFormatStringCommand(VisitorBasedCodemodCommand):
    DESCRIPTION: str = "Converts simple % style string format to f-string."

    def leave_BinaryOperation(
        self, original_node: cst.BinaryOperation, updated_node: cst.BinaryOperation
    ) -> cst.BaseExpression:
        expr_key = "expr"
        extracts = m.extract(
            original_node,
            m.BinaryOperation(
                # pyre-fixme[6]: Expected `Union[m._matcher_base.AllOf[typing.Union[m...
                left=m.MatchIfTrue(_match_simple_string),
                operator=m.Modulo(),
                # pyre-fixme[6]: Expected `Union[m._matcher_base.AllOf[typing.Union[m...
                right=m.SaveMatchedNode(
                    m.MatchIfTrue(_gen_match_simple_expression(self.module)),
                    expr_key,
                ),
            ),
        )

        if extracts:
            exprs = extracts[expr_key]
            exprs = (exprs,) if not isinstance(exprs, Sequence) else exprs
            parts = []
            simple_string = cst.ensure_type(original_node.left, cst.SimpleString)
            innards = simple_string.raw_value.replace("{", "{{").replace("}", "}}")
            tokens = innards.split("%s")
            token = tokens[0]
            if len(token) > 0:
                parts.append(cst.FormattedStringText(value=token))
            expressions: List[cst.CSTNode] = list(
                *itertools.chain(
                    [elm.value for elm in expr.elements]
                    if isinstance(expr, cst.Tuple)
                    else [expr]
                    for expr in exprs
                )
            )
            escape_transformer = EscapeStringQuote(simple_string.quote)
            i = 1
            while i < len(tokens):
                if i - 1 >= len(expressions):
                    # the %-string doesn't come with same number of elements in tuple
                    return original_node
                try:
                    parts.append(
                        cst.FormattedStringExpression(
                            expression=cast(
                                cst.BaseExpression,
                                expressions[i - 1].visit(escape_transformer),
                            )
                        )
                    )
                except Exception:
                    return original_node
                token = tokens[i]
                if len(token) > 0:
                    parts.append(cst.FormattedStringText(value=token))
                i += 1
            start = f"f{simple_string.prefix}{simple_string.quote}"
            return cst.FormattedString(
                parts=parts, start=start, end=simple_string.quote
            )

        return original_node



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from libcst._tabs import expand_tabs
from libcst.testing.utils import data_provider, UnitTest


class ExpandTabsTest(UnitTest):
    @data_provider(
        [
            ("\t", " " * 8),
            ("\t\t", " " * 16),
            ("    \t", " " * 8),
            ("\t    ", " " * 12),
            ("abcd\t", "abcd    "),
            ("abcdefg\t", "abcdefg "),
            ("abcdefgh\t", "abcdefgh        "),
            ("\tsuffix", "        suffix"),
        ]
    )
    def test_expand_tabs(self, input: str, output: str) -> None:
        self.assertEqual(expand_tabs(input), output)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
import libcst
import libcst.matchers as m
from libcst.codemod import VisitorBasedCodemodCommand


class UnnecessaryFormatString(VisitorBasedCodemodCommand):
    DESCRIPTION: str = (
        "Converts f-strings which perform no formatting to regular strings."
    )

    @m.leave(m.FormattedString(parts=(m.FormattedStringText(),)))
    def _check_formatted_string(
        self,
        _original_node: libcst.FormattedString,
        updated_node: libcst.FormattedString,
    ) -> libcst.BaseExpression:
        old_string_inner = libcst.ensure_type(
            updated_node.parts[0], libcst.FormattedStringText
        ).value
        if "{{" in old_string_inner or "}}" in old_string_inner:
            # there are only two characters we need to worry about escaping.
            return updated_node

        old_string_literal = updated_node.start + old_string_inner + updated_node.end
        new_string_literal = (
            updated_node.start.replace("f", "").replace("F", "")
            + old_string_inner
            + updated_node.end
        )

        old_string_evaled = eval(old_string_literal)  # noqa
        new_string_evaled = eval(new_string_literal)  # noqa
        if old_string_evaled != new_string_evaled:
            warn_message = (
                f"Attempted to codemod |{old_string_literal}| to "
                + f"|{new_string_literal}| but don't eval to the same! First is |{old_string_evaled}| and "
                + f"second is |{new_string_evaled}|"
            )
            self.warn(warn_message)
            return updated_node

        return libcst.SimpleString(new_string_literal)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
from pathlib import Path
from unittest.mock import Mock, patch

from libcst.metadata.full_repo_manager import FullRepoManager
from libcst.metadata.tests.test_type_inference_provider import _test_simple_class_helper
from libcst.metadata.type_inference_provider import TypeInferenceProvider
from libcst.testing.utils import UnitTest

REPO_ROOT_DIR: str = str(Path(__file__).parent.parent.parent.resolve())


class FullRepoManagerTest(UnitTest):
    @patch.object(TypeInferenceProvider, "gen_cache")
    def test_get_metadata_wrapper_with_empty_cache(self, gen_cache: Mock) -> None:
        path = "tests/pyre/simple_class.py"
        gen_cache.return_value = {path: {"types": []}}
        manager = FullRepoManager(REPO_ROOT_DIR, [path], [TypeInferenceProvider])
        wrapper = manager.get_metadata_wrapper_for_path(path)
        self.assertEqual(wrapper.resolve(TypeInferenceProvider), {})

    @patch.object(TypeInferenceProvider, "gen_cache")
    def test_get_metadata_wrapper_with_patched_cache(self, gen_cache: Mock) -> None:
        path_prefix = "tests/pyre/simple_class"
        path = f"{path_prefix}.py"
        gen_cache.return_value = {
            path: json.loads((Path(REPO_ROOT_DIR) / f"{path_prefix}.json").read_text())
        }
        manager = FullRepoManager(REPO_ROOT_DIR, [path], [TypeInferenceProvider])
        wrapper = manager.get_metadata_wrapper_for_path(path)
        _test_simple_class_helper(self, wrapper)

    @patch.object(TypeInferenceProvider, "gen_cache")
    def test_get_metadata_wrapper_with_invalid_path(self, gen_cache: Mock) -> None:
        path = "tests/pyre/simple_class.py"
        gen_cache.return_value = {path: {"types": []}}
        manager = FullRepoManager(
            REPO_ROOT_DIR, ["invalid_path.py"], [TypeInferenceProvider]
        )
        with self.assertRaisesRegex(
            Exception,
            "The path needs to be in paths parameter when constructing FullRepoManager for efficient batch processing.",
        ):
            manager.get_metadata_wrapper_for_path(path)

    @patch.object(TypeInferenceProvider, "gen_cache")
    def test_get_full_repo_cache(self, gen_cache: Mock) -> None:
        path_prefix = "tests/pyre/simple_class"
        path = f"{path_prefix}.py"
        mock_cache = {
            path: json.loads((Path(REPO_ROOT_DIR) / f"{path_prefix}.json").read_text())
        }
        gen_cache.return_value = mock_cache
        manager = FullRepoManager(REPO_ROOT_DIR, path, [TypeInferenceProvider])
        cache = manager.cache
        self.assertEqual(cache, {TypeInferenceProvider: mock_cache})



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import sys
from pathlib import Path
from subprocess import run
from typing import Iterable, List, Pattern

# Use the copyright header from this file as the benchmark for all files
EXPECTED_HEADER: str = "\n".join(
    line for line in Path(__file__).read_text().splitlines()[:4]
)

EXCEPTION_PATTERNS: List[Pattern[str]] = [
    re.compile(pattern)
    for pattern in (
        r"^native/libcst/tests/fixtures/",
        r"^libcst/_add_slots\.py$",
        r"^libcst/tests/test_(e2e|fuzz)\.py$",
        r"^libcst/_parser/base_parser\.py$",
        r"^libcst/_parser/parso/utils\.py$",
        r"^libcst/_parser/parso/pgen2/(generator|grammar_parser)\.py$",
        r"^libcst/_parser/parso/python/(py_token|tokenize)\.py$",
        r"^libcst/_parser/parso/tests/test_(fstring|tokenize|utils)\.py$",
    )
]


def tracked_files() -> Iterable[Path]:
    proc = run(
        ["git", "ls-tree", "-r", "--name-only", "HEAD"],
        check=True,
        capture_output=True,
        encoding="utf-8",
    )
    yield from (
        path
        for line in proc.stdout.splitlines()
        if not any(pattern.search(line) for pattern in EXCEPTION_PATTERNS)
        if (path := Path(line)) and path.is_file() and path.suffix in (".py", ".sh")
    )


def main() -> None:
    error = False
    for path in tracked_files():
        content = path.read_text("utf-8")
        if EXPECTED_HEADER not in content:
            print(f"Missing or incomplete copyright in {path}")
            error = True
    sys.exit(1 if error else 0)


if __name__ == "__main__":
    main()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from libcst.codemod import CodemodTest
from libcst.codemod.commands.strip_strings_from_types import StripStringsCommand


class TestStripStringsCodemod(CodemodTest):
    TRANSFORM = StripStringsCommand

    def test_noop(self) -> None:
        before = """
            foo: str = ""

            class Class:
                pass

            def foo(a: Class, **kwargs: str) -> Class:
                t: Class = Class()  # This is a comment
                bar = ""
                return t
        """
        after = """
            foo: str = ""

            class Class:
                pass

            def foo(a: Class, **kwargs: str) -> Class:
                t: Class = Class()  # This is a comment
                bar = ""
                return t
        """

        self.assertCodemod(before, after)

    def test_non_async(self) -> None:
        before = """
            class Class:
                pass

            def foo(a: "Class", **kwargs: "str") -> "Class":
                t: "Class" = Class()  # This is a comment
                return t
        """
        after = """
            from __future__ import annotations

            class Class:
                pass

            def foo(a: Class, **kwargs: str) -> Class:
                t: Class = Class()  # This is a comment
                return t
        """

        self.assertCodemod(before, after)

    def test_async(self) -> None:
        before = """
            class Class:
                pass

            async def foo(a: "Class", **kwargs: "str") -> "Class":
                t: "Class" = Class()  # This is a comment
                return t
        """
        after = """
            from __future__ import annotations

            class Class:
                pass

            async def foo(a: Class, **kwargs: str) -> Class:
                t: Class = Class()  # This is a comment
                return t
        """

        self.assertCodemod(before, after)

    def test_recursive(self) -> None:
        before = """
            class Class:
                pass

            def foo(a: List["Class"]):
                pass

            def bar(a: List[Optional["Class"]]):
                pass

            def baz(a: "List[Class]"):
                pass
        """
        after = """
            from __future__ import annotations

            class Class:
                pass

            def foo(a: List[Class]):
                pass

            def bar(a: List[Optional[Class]]):
                pass

            def baz(a: List[Class]):
                pass
        """

        self.assertCodemod(before, after)

    def test_literal(self) -> None:
        before = """
            from typing_extensions import Literal

            class Class:
                pass

            def foo(a: Literal["one", "two", "three"]):
                pass

            def bar(a: Union["Class", Literal["one", "two", "three"]]):
                pass
        """
        after = """
            from __future__ import annotations
            from typing_extensions import Literal

            class Class:
                pass

            def foo(a: Literal["one", "two", "three"]):
                pass

            def bar(a: Union[Class, Literal["one", "two", "three"]]):
                pass
        """

        self.assertCodemod(before, after)

    def test_literal_alias(self) -> None:
        before = """
            from typing_extensions import Literal as Lit

            class Class:
                pass

            def foo(a: Lit["one", "two", "three"]):
                pass

            def bar(a: Union["Class", Lit["one", "two", "three"]]):
                pass
        """
        after = """
            from __future__ import annotations
            from typing_extensions import Literal as Lit

            class Class:
                pass

            def foo(a: Lit["one", "two", "three"]):
                pass

            def bar(a: Union[Class, Lit["one", "two", "three"]]):
                pass
        """

        self.assertCodemod(before, after)

    def test_literal_object(self) -> None:
        before = """
            import typing_extensions

            class Class:
                pass

            def foo(a: typing_extensions.Literal["one", "two", "three"]):
                pass

            def bar(a: Union["Class", typing_extensions.Literal["one", "two", "three"]]):
                pass
        """
        after = """
            from __future__ import annotations
            import typing_extensions

            class Class:
                pass

            def foo(a: typing_extensions.Literal["one", "two", "three"]):
                pass

            def bar(a: Union[Class, typing_extensions.Literal["one", "two", "three"]]):
                pass
        """

        self.assertCodemod(before, after)

    def test_literal_object_alias(self) -> None:
        before = """
            import typing_extensions as typext

            class Class:
                pass

            def foo(a: typext.Literal["one", "two", "three"]):
                pass

            def bar(a: Union["Class", typext.Literal["one", "two", "three"]]):
                pass
        """
        after = """
            from __future__ import annotations
            import typing_extensions as typext

            class Class:
                pass

            def foo(a: typext.Literal["one", "two", "three"]):
                pass

            def bar(a: Union[Class, typext.Literal["one", "two", "three"]]):
                pass
        """

        self.assertCodemod(before, after)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Dict, Sequence, Union

import libcst
import libcst.matchers as m
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.helpers import insert_header_comments


class FixPyreDirectivesCommand(VisitorBasedCodemodCommand):
    """
    Given a source file, we'll move the any strict or unsafe tag to the top of the
    file if it contains one. Also tries to fix typo'd directives.
    """

    DESCRIPTION: str = "Fixes common misspelling and location errors with pyre tags."

    PYRE_TAGS: Sequence[str] = ["strict", "unsafe"]

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self.move_strict: Dict[str, bool] = {tag: False for tag in self.PYRE_TAGS}
        self.module_header_tags: Dict[str, int] = {tag: 0 for tag in self.PYRE_TAGS}
        self.in_module_header: bool = False

    def visit_Module_header(self, node: libcst.Module) -> None:
        if self.in_module_header:
            raise Exception("Logic error!")
        self.in_module_header = True

    def leave_Module_header(self, node: libcst.Module) -> None:
        if not self.in_module_header:
            raise Exception("Logic error!")
        self.in_module_header = False

    def leave_EmptyLine(
        self, original_node: libcst.EmptyLine, updated_node: libcst.EmptyLine
    ) -> Union[libcst.EmptyLine, libcst.RemovalSentinel]:
        # First, find misplaced lines.
        for tag in self.PYRE_TAGS:
            if m.matches(updated_node, m.EmptyLine(comment=m.Comment(f"# pyre-{tag}"))):
                if self.in_module_header:
                    # We only want to remove this if we've already found another
                    # pyre-strict in the header (that means its duplicated). We
                    # also don't want to move the pyre-strict since its already in
                    # the header, so don't mark that we need to move.
                    self.module_header_tags[tag] += 1
                    if self.module_header_tags[tag] > 1:
                        return libcst.RemoveFromParent()
                    else:
                        return updated_node
                else:
                    # This showed up outside the module header, so move it inside
                    if self.module_header_tags[tag] < 1:
                        self.move_strict[tag] = True
                    return libcst.RemoveFromParent()
            # Now, find misnamed lines
            if m.matches(updated_node, m.EmptyLine(comment=m.Comment(f"# pyre {tag}"))):
                if self.in_module_header:
                    # We only want to remove this if we've already found another
                    # pyre-strict in the header (that means its duplicated). We
                    # also don't want to move the pyre-strict since its already in
                    # the header, so don't mark that we need to move.
                    self.module_header_tags[tag] += 1
                    if self.module_header_tags[tag] > 1:
                        return libcst.RemoveFromParent()
                    else:
                        return updated_node.with_changes(
                            comment=libcst.Comment(f"# pyre-{tag}")
                        )
                else:
                    # We found an intended pyre-strict, but its spelled wrong. So, remove it
                    # and re-add a new one in leave_Module.
                    if self.module_header_tags[tag] < 1:
                        self.move_strict[tag] = True
                    return libcst.RemoveFromParent()
        # We found a regular comment, don't care about this.
        return updated_node

    def leave_Module(
        self, original_node: libcst.Module, updated_node: libcst.Module
    ) -> libcst.Module:
        comments = [f"# pyre-{tag}" for tag in self.PYRE_TAGS if self.move_strict[tag]]
        return insert_header_comments(updated_node, comments)
